import React, { useEffect, useRef, useState } from 'react'
import PropTypes from 'prop-types'
import { ChevronDown } from '@kyper/icon/ChevronDown'
import { ChevronUp } from '@kyper/icon/ChevronUp'
import { useTokens } from '@kyper/tokenprovider'
import { css } from '@mxenabled/cssinjs'
import { useCombobox } from 'downshift'
import { getFilteredItems, itemShape } from './lib/shared/utils'
import { DescriptionText } from './lib/shared/DescriptionText'
import { ErrorText } from './lib/shared/ErrorText'
import { SelectDropDown } from './lib/shared/SelectDropDown'

const INPUT_HEIGHT = 48
const LABEL_LENGTH = 35

const getInitialSelectedItemValue = initialSelectedItem => {
  if (typeof initialSelectedItem?.label !== 'string' && initialSelectedItem?.readableLabel) {
    return initialSelectedItem.readableLabel
  } else {
    return initialSelectedItem?.label
  }
}

export const Combobox = ({
  ariaLabel = null,
  autoreset = true,
  className = '',
  description,
  dropDownProps,
  errorText,
  initialSelectedItem,
  isDisabled = false,
  items,
  label,
  name,
  onChange,
  placeholder = 'Select a value',
  showSecondaryLabelBackground,
  showDescriptionIcon,
  showErrorIcon,
  style,
}) => {
  const [ariaDescribedBy, setAriaDescribedBy] = useState(null)
  const [inputValue, setInputValue] = useState(
    initialSelectedItem ? getInitialSelectedItemValue(initialSelectedItem) : '',
  )
  const [showAllItems, setShowAllItems] = useState(true)
  const filteredItems = getFilteredItems(items, showAllItems, inputValue)
  const itemToString = item => (item ? item.label : '')

  const tokens = useTokens()
  // this allows for multiline labels
  const labelPosition =
    label?.length > LABEL_LENGTH ? -tokens.Spacing.Medium : -tokens.Spacing.XSmall
  const comboStyles = getComboStyles(tokens, labelPosition, label, showSecondaryLabelBackground)

  // downshift
  const {
    closeMenu,
    getToggleButtonProps,
    getLabelProps,
    getMenuProps,
    getInputProps,
    getComboboxProps,
    getItemProps,
    highlightedIndex,
    isOpen,
    reset,
    toggleMenu,
  } = useCombobox({
    initialSelectedItem,
    initialInputValue: getInitialSelectedItemValue(initialSelectedItem),
    inputValue,
    items: filteredItems,
    itemToString,
    stateReducer: (_, actionAndChanges) => {
      const { changes, type } = actionAndChanges

      switch (type) {
        case useCombobox.stateChangeTypes.InputKeyDownEnter:
        case useCombobox.stateChangeTypes.ItemClick:
          return {
            ...changes,
            isOpen: true,
          }
      }
      return changes
    },
    onStateChange: ({ inputValue, type, selectedItem, isOpen }) => {
      switch (type) {
        case useCombobox.stateChangeTypes.InputKeyDownEscape:
          setInputValue('')
          break
        case useCombobox.stateChangeTypes.InputChange:
          setShowAllItems(false)
          setInputValue(inputValue)
          break
        case useCombobox.stateChangeTypes.ToggleButtonClick:
          if (isOpen) setShowAllItems(true)
          break
        case useCombobox.stateChangeTypes.InputKeyDownEnter:
        case useCombobox.stateChangeTypes.ItemClick:
          if (selectedItem) {
            // uses the readableLabel if the label is a jsx component
            setInputValue(
              typeof selectedItem.label === 'object'
                ? selectedItem.readableLabel ?? selectedItem.value
                : selectedItem.readableLabel ?? selectedItem.label,
            )
            const currentSelectedItem = {
              ...selectedItem,
              // add fake target object properties for validation
              target: {
                name,
                value: selectedItem.value,
              },
            }

            onChange(currentSelectedItem)
          }
          closeMenu()
          break
        default:
          break
      }
    },
  })

  const inputProps = getInputProps({
    disabled: isDisabled,
    name,
    onChange,
    onClick() {
      if (!isOpen) {
        toggleMenu()
      }
    },
    placeholder,
    type: 'text',
  })
  const buttonProps = getToggleButtonProps({
    disabled: isDisabled,
    'aria-labelledby': getLabelProps().id,
  })

  //side effects
  useEffect(() => {
    if (errorText) {
      setAriaDescribedBy(`${inputProps.id}-error`)
    } else if (description) {
      setAriaDescribedBy(`${inputProps.id}-description`)
    } else {
      setAriaDescribedBy(null)
    }
  }, [errorText, inputProps.id, description])

  const values = JSON.stringify(items.map(item => item.value))
  const valuesRef = useRef(values)

  useEffect(() => {
    const prevValues = valuesRef.current

    /**
     * In order to watch the items list for any changes, and to be able to autoreset, we
     * have to be able to compare the llist with a previous state. Since the item.label
     * can be a string or a node type, we can't stringify the entire array of items
     * because node types can't be converted to JSON. So instead, we will compare the
     * list of `item.values` since values should always be a type that can be converted
     * to JSON
     */
    if (autoreset && values && values !== prevValues) {
      valuesRef.current = items
      reset()
      setInputValue('')
    }
  }, [autoreset, values])

  return (
    <div
      className={`${css(comboStyles.container)} ${isDisabled ? 'disabled' : ''} ${
        errorText ? 'error' : ''
      } kyper-combobox ${className}`}
      style={style}
    >
      <label
        {...getLabelProps()}
        className={css({
          ...comboStyles.label,
          ...(!label && {
            position: 'absolute',
            left: -10000,
          }),
        })}
      >
        {label || ariaLabel}
      </label>
      <div {...getComboboxProps()} className={css({ position: 'relative' })}>
        <input
          aria-describedby={ariaDescribedBy}
          aria-invalid={!!errorText}
          {...inputProps}
          className={css(comboStyles.input)}
        />
        <button {...buttonProps} className={css(comboStyles.toggleButton)}>
          {isOpen ? <ChevronUp /> : <ChevronDown />}
        </button>
      </div>
      <SelectDropDown
        ariaLabelledBy={`${buttonProps['aria-labelledby']}`}
        getItemProps={getItemProps}
        getMenuProps={getMenuProps}
        highlightedIndex={highlightedIndex}
        isOpen={isOpen}
        items={filteredItems}
        {...dropDownProps}
      />
      {errorText ? (
        <ErrorText
          buttonPropsId={inputProps.id}
          errorText={errorText}
          showErrorIcon={showErrorIcon}
        />
      ) : null}

      {description && !errorText ? (
        <DescriptionText
          buttonPropsId={inputProps.id}
          description={description}
          showDescriptionIcon={showDescriptionIcon}
        />
      ) : null}
    </div>
  )
}

const getComboStyles = (tokens, labelPosition, label, showSecondaryLabelBackground) => ({
  container: {
    width: '100%',
    boxSizing: 'border-box',
    position: 'relative',
    fontFamily: tokens.Font.Regular,
    '&.disabled': {
      cursor: 'not-allowed',
    },
    '&.disabled input': {
      backgroundColor: `${tokens.BackgroundColor.InputDisabled} !important`,
      borderColor: `${tokens.BorderColor.InputDisabled} !important`,
      pointerEvents: 'none',
      color: `${tokens.TextColor.Disabled} !important`,
    },
    '&.disabled button': {
      color: tokens.TextColor.Disabled,
      cursor: 'not-allowed',
    },
    '&.disabled label': {
      color: `${tokens.TextColor.Disabled} !important`,
    },
    '&.error input': {
      border: `1px solid ${tokens.BorderColor.InputError} !important`,
      ':focus': {
        boxShadow: `inset 0px 0px 0px 2px ${tokens.BorderColor.InputError} !important`,
      },
    },
    '&.error label': {
      color: `${tokens.TextColor.Error} !important`,
    },
  },
  input: {
    width: '100%',
    height: INPUT_HEIGHT,
    boxSizing: 'border-box',
    boxShadow: `inset 0px 0px 0px 0px ${tokens.BorderColor.InputHover}`,
    color: tokens.TextColor.Default,
    borderRadius: tokens.BorderRadius.Input,
    border: `1px solid ${tokens.BorderColor.Input}`,
    backgroundColor: 'transparent',
    fontSize: tokens.FontSize.Input,
    transition: `box-shadow 350ms ${tokens.Easing.Default}`,
    // Top is the only unique value when there is a label.
    paddingTop: label ? tokens.Spacing.InputPaddingTop : tokens.Spacing.InputPaddingLeft,
    paddingRight: tokens.Spacing.XXLarge,
    paddingBottom: tokens.Spacing.InputPaddingLeft,
    paddingLeft: tokens.Spacing.InputPaddingLeft,
    lineHeight: tokens.LineHeight.Body,
    '&:focus': {
      outline: 'none',
      border: `1px solid ${tokens.BorderColor.InputFocus}`,
      boxShadow: `inset 0px 0px 0px 2px ${tokens.BorderColor.InputFocus} !important`,
    },
    '&:hover': {
      boxShadow: `inset 0px 0px 0px 1px ${tokens.BorderColor.InputHover}`,
    },
    '&::placeholder': {
      color: tokens.TextColor.InputPlaceholder,
    },
  },
  label: {
    position: 'absolute',
    top: labelPosition,
    left: tokens.Spacing.InputLabelMarginLeft,
    fontSize: tokens.FontSize.InputLabel,
    paddingLeft: tokens.Spacing.InputLabelPaddingLeft,
    paddingRight: tokens.Spacing.InputLabelPaddingRight,
    marginRight: tokens.Spacing.Small,
    backgroundColor: showSecondaryLabelBackground
      ? tokens.BackgroundColor.InputLabelSecondary
      : tokens.BackgroundColor.InputLabelDefault,
    color: tokens.TextColor.InputLabel,
    lineHeight: tokens.LineHeight.Small,
    zIndex: 1,
  },
  toggleButton: {
    position: 'absolute',
    display: 'flex',
    alignItems: 'center',
    justifyContent: 'center',
    top: 0,
    right: 0,
    width: 44,
    height: 48,
    border: 'none',
    background: 'transparent',
    cursor: 'pointer',
    color: tokens.TextColor.Default,
    '&:focus': {
      outline: 'none',
    },
  },
})

Combobox.propTypes = {
  /**
   * When you don't pass in a visible label you need to pass an ariaLabel so it's accessible for screen-readers
   */
  ariaLabel: PropTypes.string,
  /**
   * ariaLabelCheck throws a console error if you forget to pass in an aria-label when not
   * passing in a label for screen-readers
   */
  ariaLabelCheck: (props, _, componentName) => {
    /* eslint-disable object-shorthand */
    if (!props['label'] && !props['ariaLabel']) {
      return new Error(
        `For screenreaders you must use the ariaLabel prop when not using a label for the ${componentName} component.`,
      )
    }
    return null
  },
  /**
   * By default when a Combobox items change the selected item will reset automatically. To change this set autoreset to false.
   */
  autoreset: PropTypes.bool,
  className: PropTypes.string,
  /**
   * Text to show description below the Combobox when you need more info than just the label.
   */
  description: PropTypes.string,
  /**
   * props to pass down to the Combobox dropdown.
   */
  dropDownProps: PropTypes.object,
  /**
   * Text to show an error below the Combobox when there is a validation error.
   */
  errorText: PropTypes.string,
  /**
   * The item you want selected when the component mounts
   */
  initialSelectedItem: itemShape,
  /**
   * Disables the Combobox when true
   */
  isDisabled: PropTypes.bool,
  /**
   * An array of the items that can be selected.
   *
   * `label` and `value` are always required. Pass in a `readableLabel` when you pass a custom jsx element in the `label`. The `readableLabel` is displayed in the search input when you select an item.
   *

    [{
      label: string | element,
      readableLabel: string,
      value: string | number
    }]

   */
  items: PropTypes.arrayOf(itemShape).isRequired,
  label: PropTypes.string,
  /**
   * Used for validation. Returned in the selectedItem object to the onChange function
   */
  name: PropTypes.string,
  /**
   * Function to handle item selection changes
   */
  onChange: PropTypes.func.isRequired,
  placeholder: PropTypes.string,
  /**
   * Show an icon next to the description text
   */
  showDescriptionIcon: PropTypes.bool,
  /**
   * Show an icon next to the error text
   */
  showErrorIcon: PropTypes.bool,
  /**
   * Use when your label needs a background for color contrast
   */
  showSecondaryLabelBackground: PropTypes.bool,
  style: PropTypes.object,
}
