import React, { useEffect, useState, useRef } from 'react'
import { useSelect } from 'downshift'
import PropTypes from 'prop-types'
import { css } from '@mxenabled/cssinjs'
import { useTokens } from '@kyper/tokenprovider'
import { ChevronDown } from '@kyper/icon/ChevronDown'
import { ChevronUp } from '@kyper/icon/ChevronUp'
import { DescriptionText } from './lib/shared/DescriptionText'
import { ErrorText } from './lib/shared/ErrorText'
import { SelectDropDown } from './lib/shared/SelectDropDown'

const LABEL_LENGTH = 35

const getSelectedItemIndex = (currentSelectedItem, items) =>
  items.findIndex(item => {
    return item.value === currentSelectedItem?.value
  })

export const Select = ({
  ariaLabel = null,
  autoreset = true,
  className = '',
  description,
  dropDownProps,
  errorText,
  initialSelectedItem,
  isDisabled = false,
  items,
  label,
  name,
  onChange,
  placeholder = 'Select a value',
  showSecondaryLabelBackground,
  showDescriptionIcon,
  showErrorIcon,
  style,
}) => {
  const {
    isOpen,
    selectedItem,
    getToggleButtonProps,
    getLabelProps,
    getMenuProps,
    highlightedIndex,
    getItemProps,
    reset,
  } = useSelect({
    items,
    initialSelectedItem,
    selectedItem: initialSelectedItem,
    itemToString: item => (item === null ? '' : String(item.label)),
    onSelectedItemChange: ({ selectedItem }) => {
      const currentSelectedItem = { ...selectedItem }

      if (selectedItem) {
        currentSelectedItem.target = {}
        currentSelectedItem.target.name = name
        currentSelectedItem.target.value = selectedItem.value
      }

      onChange(currentSelectedItem)
      setCurrentSelectedItem(selectedItem)
      setCurrentSelectedItemIndex(getSelectedItemIndex(selectedItem, items))
    },
  })
  const tokens = useTokens()
  // this allows for multiline labels
  const labelPosition =
    label?.length > LABEL_LENGTH ? -tokens.Spacing.Medium : -tokens.Spacing.XSmall
  const styles = getStyles(tokens, labelPosition, label, showSecondaryLabelBackground)

  const buttonProps = getToggleButtonProps({
    disabled: isDisabled,
    name,
  })

  const [ariaDescribedBy, setAriaDescribedBy] = useState(null)
  const [currentSelectedItem, setCurrentSelectedItem] = useState(initialSelectedItem)
  const [currentSelectedItemIndex, setCurrentSelectedItemIndex] = useState(
    getSelectedItemIndex(initialSelectedItem, items),
  )

  useEffect(() => {
    if (errorText) {
      setAriaDescribedBy(`${buttonProps.id}-error`)
    } else if (description) {
      setAriaDescribedBy(`${buttonProps.id}-description`)
    } else {
      setAriaDescribedBy(null)
    }
  }, [errorText, buttonProps.id, description])

  const values = JSON.stringify(items.map(item => item.value))
  const valuesRef = useRef(values)

  useEffect(() => {
    const prevValues = valuesRef.current

    /**
     * In order to watch the items list for any changes, and to be able to autoreset, we
     * have to be able to compare the list with a previous state. Since the item.label
     * can be a string or a node type, we can't stringify the entire array of items
     * because node types can't be converted to JSON. So instead, we will compare the
     * list of `item.values` since values should always be a type that can be converted
     * to JSON
     */
    if (autoreset && values && values !== prevValues) {
      valuesRef.current = items
      reset()
    }
  }, [autoreset, values])

  return (
    <div
      className={`${css(styles.container)} ${isDisabled ? 'disabled' : ''} ${
        errorText ? 'error' : ''
      } kyper-select ${className}`}
      style={style}
    >
      <label
        aria-label={ariaLabel}
        {...getLabelProps()}
        className={css({
          ...styles.label,
          ...(!label && {
            position: 'absolute',
            left: -10000,
          }),
        })}
      >
        {label}
      </label>
      <button
        aria-invalid={!!errorText}
        className={`${css(styles.button)} ${!selectedItem ? 'placeholder' : ''}`}
        {...buttonProps}
        aria-describedby={ariaDescribedBy}
        type="button"
      >
        <span>{(currentSelectedItem && currentSelectedItem.label) || placeholder}</span>
        <span className={css(styles.iconContainer)}>
          {isOpen ? <ChevronUp /> : <ChevronDown />}
        </span>
      </button>
      <SelectDropDown
        ariaLabelledBy={`${buttonProps['aria-labelledby']}`}
        currentSelectedItemIndex={currentSelectedItemIndex}
        getItemProps={getItemProps}
        getMenuProps={getMenuProps}
        highlightedIndex={highlightedIndex}
        isOpen={isOpen}
        items={items}
        {...dropDownProps}
      />
      {errorText ? (
        <ErrorText
          buttonPropsId={buttonProps.id}
          errorText={errorText}
          showErrorIcon={showErrorIcon}
        />
      ) : null}

      {description && !errorText ? (
        <DescriptionText
          buttonPropsId={buttonProps.id}
          description={description}
          showDescriptionIcon={showDescriptionIcon}
        />
      ) : null}
    </div>
  )
}

const getStyles = (tokens, labelPosition, label, showSecondaryLabelBackground) => {
  return {
    container: {
      '& p': {
        marginBottom: 0,
      },
      width: '100%',
      position: 'relative',
      fontFamily: tokens.Font.Regular,
      boxSizing: 'border-box',
      '&.disabled': {
        cursor: 'not-allowed',
      },
      '&.disabled button': {
        backgroundColor: tokens.BackgroundColor.InputDisabled,
        borderColor: tokens.BorderColor.InputDisabled,
        pointerEvents: 'none',
        color: tokens.TextColor.Disabled,
      },
      '&.disabled label': {
        color: `${tokens.TextColor.Disabled} !important`,
      },
      '&.error button': {
        border: `1px solid ${tokens.BorderColor.InputError} !important`,
        ':focus': {
          boxShadow: `inset 0px 0px 0px 2px ${tokens.BorderColor.InputError} !important`,
        },
      },
      '&.error label': {
        color: `${tokens.TextColor.Error} !important`,
      },
    },
    button: {
      width: '100%',
      height: 48,
      boxSizing: 'border-box',
      textAlign: 'left',
      color: tokens.TextColor.Default,
      borderRadius: tokens.BorderRadius.Input,
      border: `1px solid ${tokens.BorderColor.Input}`,
      backgroundColor: 'transparent',
      paddingLeft: tokens.Spacing.InputPaddingLeft,
      paddingTop: label ? tokens.Spacing.InputPaddingTop : tokens.Spacing.InputPaddingLeft,
      paddingRight: 36,
      paddingBottom: tokens.Spacing.InputPaddingBottom,
      fontSize: tokens.FontSize.Input,
      lineHeight: tokens.LineHeight.Body,
      transition: `box-shadow ${tokens.Time.Short} ${tokens.Easing.Default}`,
      '&:focus': {
        outline: 'none',
        border: `1px solid ${tokens.BorderColor.InputFocus}`,
        boxShadow: `inset 0px 0px 0px 2px ${tokens.BorderColor.InputFocus} !important`,
      },
      '&:hover': {
        boxShadow: `inset 0px 0px 0px 1px ${tokens.BorderColor.InputHover}`,
      },
      '&.placeholder': {
        color: tokens.TextColor.InputPlaceholder,
      },
      // Button text
      '& span:first-of-type': {
        display: 'block',
        overflow: 'hidden',
        whiteSpace: 'nowrap',
        textOverflow: 'ellipsis',
      },
    },
    label: {
      position: 'absolute',
      top: labelPosition,
      left: tokens.Spacing.InputLabelMarginLeft,
      fontSize: tokens.FontSize.InputLabel,
      paddingLeft: tokens.Spacing.InputLabelPaddingLeft,
      paddingRight: tokens.Spacing.InputLabelPaddingRight,
      marginRight: tokens.Spacing.Small,
      backgroundColor: showSecondaryLabelBackground
        ? tokens.BackgroundColor.InputLabelSecondary
        : tokens.BackgroundColor.InputLabelDefault,
      color: tokens.TextColor.InputLabel,
      lineHeight: tokens.LineHeight.Small,
    },
    iconContainer: {
      position: 'absolute',
      top: label ? 18 : tokens.Spacing.Medium,
      right: 12,
      left: 'auto',
      display: 'flex',
      flexDirection: 'column',
    },
  }
}

Select.propTypes = {
  /**
   * When you don't pass in a visible label you need to pass an ariaLabel so it's accessible for screen-readers
   */
  ariaLabel: PropTypes.string,
  /**
   * ariaLabelCheck throws a console error if you forget to pass in an ariaLabel when not
   * passing in a visible label for screen-readers
   */
  ariaLabelCheck: (props, _, componentName) => {
    /* eslint-disable object-shorthand */
    if (!props['label'] && !props['ariaLabel']) {
      return new Error(
        `For screenreaders you must use the ariaLabel prop when not using a label for the ${componentName} component.`,
      )
    }
    return null
  },
  /**
   * By default when a Selects items change the selected item will reset automatically. To change this set autoreset to false.
   */
  autoreset: PropTypes.bool,
  className: PropTypes.string,
  /**
   * Text to show description below the Select when you need more info than just the label.
   */
  description: PropTypes.string,
  /**
   * props to pass down to the Select dropdown.
   */
  dropDownProps: PropTypes.object,
  /**
   * Text to show an error below the Select when there is a validation error.
   */
  errorText: PropTypes.string,
  /**
   * The item you want selected when the component mounts
   */
  initialSelectedItem: PropTypes.shape({
    label: PropTypes.string.isRequired,
    value: PropTypes.string.isRequired,
  }),
  /**
   * Disables the Combobox when true
   */
  isDisabled: PropTypes.bool,
  /**
   * An array of the items that can be selected
   */
  items: PropTypes.arrayOf(
    PropTypes.shape({
      label: PropTypes.oneOfType([PropTypes.string, PropTypes.node]).isRequired,
      value: PropTypes.string.isRequired,
    }),
  ).isRequired,
  label: PropTypes.string,
  /**
   * Used for validation. Returned in the selectedItem object to the onChange function
   */
  name: PropTypes.string,
  /**
   * Function to handle item selection changes
   */
  onChange: PropTypes.func.isRequired,
  placeholder: PropTypes.string,
  /**
   * Show an icon next to the description text
   */
  showDescriptionIcon: PropTypes.bool,
  /**
   * Show an icon next to the error text
   */
  showErrorIcon: PropTypes.bool,
  /**
   * Use when your label needs a background for color contrast
   */
  showSecondaryLabelBackground: PropTypes.bool,
  style: PropTypes.object,
}
