import { useTokens } from '@kyper/tokenprovider'
import { css } from '@mxenabled/cssinjs'
import PropTypes from 'prop-types'
import React from 'react'
import { itemShape } from './utils'

const PLACEMENT = {
  LEFT: 'left',
  RIGHT: 'right',
}

const POSITION = {
  ABSOLUTE: 'absolute',
  RELATIVE: 'relative',
}

const propTypes = {
  ariaLabelledBy: PropTypes.string,
  className: PropTypes.string,
  currentSelectedItemIndex: PropTypes.number,
  getItemProps: PropTypes.func.isRequired,
  getMenuProps: PropTypes.func.isRequired,
  highlightedIndex: PropTypes.number,
  isOpen: PropTypes.bool,
  isScrollable: PropTypes.bool,
  items: PropTypes.arrayOf(itemShape).isRequired,
  ListItemComponent: PropTypes.element,
  placement: PropTypes.oneOf([PLACEMENT.LEFT, PLACEMENT.RIGHT]),
  position: PropTypes.oneOf([POSITION.ABSOLUTE, POSITION.RELATIVE]),
  width: PropTypes.number,
}

export const SelectDropDown = ({
  ariaLabelledBy,
  className = '',
  currentSelectedItemIndex,
  getItemProps,
  getMenuProps,
  highlightedIndex,
  isOpen,
  isScrollable = true,
  items,
  ListItemComponent = null,
  placement = PLACEMENT.LEFT,
  position = POSITION.ABSOLUTE,
  width,
  ...rest
}) => {
  const tokens = useTokens()
  const styles = getDropdownStyle(tokens, isOpen, items, width)
  const menuProps = getMenuProps({
    'aria-labelledby': ariaLabelledBy,
  })

  return (
    <ul
      className={`${css(styles.container)} ${className} ${
        isScrollable ? 'scrollable' : ''
      } ${position} ${placement}`}
      {...menuProps}
      {...rest}
    >
      {isOpen
        ? items.map((item, index) => {
            const itemHighlighted = highlightedIndex === index || currentSelectedItemIndex === index

            return !item.doNotAllowSelection ? (
              <li
                className={css({
                  ...styles.item,
                  ...(itemHighlighted && { ...styles.itemHighlighted }),
                })}
                key={`${item.label}-${index}`}
                role="option"
                {...getItemProps({
                  item: item.label,
                  index,
                })}
              >
                {ListItemComponent ? (
                  <ListItemComponent label={item.label} value={item.value} />
                ) : (
                  item.label
                )}
              </li>
            ) : (
              <span
                className={css({ ...styles.item, cursor: 'default' })}
                key={`${item.label}${index}`}
              >
                {item.label}
              </span>
            )
          })
        : null}
    </ul>
  )
}

const getDropdownStyle = (tokens, isOpen, items, width) => ({
  container: {
    zIndex: 100,
    position: 'relative',
    color: tokens.TextColor.Default,
    boxSizing: 'border-box',
    boxShadow: tokens.BoxShadow.DropdownMenu,
    borderRadius: tokens.BorderRadius.DropdownMenu,
    fontSize: tokens.FontSize.DropdownMenu,
    backgroundColor: tokens.BackgroundColor.DropdownMenu,
    marginTop: tokens.Spacing.Tiny,
    padding: items.length && isOpen ? tokens.Spacing.Tiny : 0,
    width: width ? width : '100%',
    '&:focus': {
      outline: 'none',
    },
    '&.scrollable': {
      maxHeight: 300,
      overflow: 'auto',
    },
    '&.absolute': {
      position: 'absolute',
    },
    '&.left': {
      left: 0,
    },
    '&.right': {
      right: 0,
    },
  },
  item: {
    display: 'flex',
    gap: tokens.Spacing.Small,
    alignItems: 'center',
    listStyle: 'none',
    cursor: 'pointer',
    paddingTop: tokens.Spacing.Small,
    paddingRight: tokens.Spacing.XSmall,
    paddingBottom: tokens.Spacing.Small,
    paddingLeft: tokens.Spacing.XSmall,
    borderRadius: tokens.BorderRadius.DropdownMenu,
    fontSize: tokens.FontSize.DropdownItem,
    backgroundColor: tokens.BackgroundColor.DropdownMenu,
    marginBottom: tokens.Spacing.Tiny,
    wordBreak: 'break-word',
    '&:last-child': {
      marginBottom: 0,
    },
  },
  itemHighlighted: {
    backgroundColor: tokens.BackgroundColor.DropdownMenuItemActive,
    color: tokens.TextColor.DropdownMenuItemActive,
  },
})

SelectDropDown.propTypes = propTypes
